#include <types.h>

__global__ void collision(TYPE* gpu_c1, TYPE* gpu_c2, TYPE* gpu_c3,
			  TYPE* gpu_c4, TYPE* gpu_c5, TYPE* gpu_c6,
			  TYPE* gpu_k1, TYPE* gpu_k2, TYPE* gpu_k3,
			  TYPE* gpu_k4, TYPE* gpu_k5, TYPE* gpu_k6,
			  TYPE* gpu_boundary) {
	unsigned int ro;
	int x = blockIdx.x * BLOCK_DIM + threadIdx.x; 
	int y = blockIdx.y * BLOCK_DIM + threadIdx.y; 
	int index = y * LATTICE_WIDTH + x;

	if(x > LATTICE_WIDTH || y > LATTICE_WIDTH || index > (LATTICE_WIDTH * LATTICE_HEIGHT)) {
		return;
	}


	TYPE a = gpu_c1[index];
	TYPE b = gpu_c2[index];
	TYPE c = gpu_c3[index];
	TYPE d = gpu_c4[index];
	TYPE e = gpu_c5[index];
	TYPE f = gpu_c6[index];

	TYPE tripple = (a ^ b) & (b ^ c) & (c ^ d) & (d ^ e) & (e ^ f);

	TYPE db1  = (a & d & ~(b | c | e | f));
	TYPE db2  = (b & e & ~(a | c | d | f));
	TYPE db3  = (c & f & ~(a | b | d | e));

	ro        = (index  + x* 32131) % 2; //flip_coin( );

	int right_boundary = (x + 1 < LATTICE_WIDTH)? gpu_boundary[index + 1] : 1;
	int left_boundary  = (x - 1 > 0)? gpu_boundary[index - 1] : 1;
	int up_boundary    = (y + 1 < LATTICE_HEIGHT)? gpu_boundary[index + LATTICE_WIDTH] : 1;
	int down_boundary  = (y - 1 > 0)? gpu_boundary[index - LATTICE_WIDTH] : 1;
	
	int boundary = right_boundary | left_boundary | up_boundary | down_boundary;

	gpu_k1[index] = boundary & (a ^ (((tripple | db1 | (ro & db2) | (~ro & db3))) & gpu_boundary[index])) | (~boundary & d);
	gpu_k2[index] = boundary & (b ^ (((tripple | db2 | (ro & db3) | (~ro & db1))) & gpu_boundary[index])) | (~boundary & e);
	gpu_k3[index] = boundary & (c ^ (((tripple | db3 | (ro & db1) | (~ro & db2))) & gpu_boundary[index])) | (~boundary & f);
	gpu_k4[index] = boundary & (d ^ (((tripple | db1 | (ro & db2) | (~ro & db3))) & gpu_boundary[index])) | (~boundary & a);
	gpu_k5[index] = boundary & (e ^ (((tripple | db2 | (ro & db3) | (~ro & db1))) & gpu_boundary[index])) | (~boundary & b);
	gpu_k6[index] = boundary & (f ^ (((tripple | db3 | (ro & db1) | (~ro & db2))) & gpu_boundary[index])) | (~boundary & c);
}

__global__ void propagate(TYPE* gpu_c1, TYPE* gpu_c2, TYPE* gpu_c3,
			  TYPE* gpu_c4, TYPE* gpu_c5, TYPE* gpu_c6,
			  TYPE* gpu_k1, TYPE* gpu_k2, TYPE* gpu_k3,
			  TYPE* gpu_k4, TYPE* gpu_k5, TYPE* gpu_k6,
			  TYPE* gpu_boundary) {
	int x = blockIdx.x * BLOCK_DIM + threadIdx.x; 
	int y = blockIdx.y  * BLOCK_DIM + threadIdx.y; 
	int index = y * LATTICE_WIDTH + x;

	if(index > (LATTICE_WIDTH * LATTICE_HEIGHT)) {
		return;
	}
	
	int right = (x+1) % LATTICE_WIDTH;
	int left = (x - 1 + LATTICE_WIDTH) % LATTICE_WIDTH;
	int up = (y+1) % LATTICE_HEIGHT;
	int down = (y-1 + LATTICE_HEIGHT) % LATTICE_HEIGHT;
	
	gpu_c1[y  * LATTICE_WIDTH + right] = gpu_k1[index];
	gpu_c4[y  * LATTICE_WIDTH + left] = gpu_k4[index];

	gpu_c2[up * LATTICE_WIDTH + x] = gpu_k2[index];
	gpu_c5[down* LATTICE_WIDTH + x] = gpu_k5[index];

	gpu_c3[up * LATTICE_WIDTH + left] = gpu_k3[index];
	gpu_c6[down* LATTICE_WIDTH + right] = gpu_k6[index];
}
